#include <stdio.h>
#include <stdlib.h>
#include "genetic.h"
#include "rand32.h"


// New implementation

unsigned int genetic(int Np, int Ng, double Rc, double Rm, int max_init_fitness) {
  unsigned int * population;
  int          * fitnesses;
  bool         * alive;

  int psize = (int) ((1.0 + Rc) * Np);
  int nkids = psize - Np;
  int i, a, b, kid,sum,running_sum;
  unsigned int reduction_or, reduction_and;
  reduction_or = 0;
  reduction_and = 0xFFFFFFFF;

  unsigned int best;
  int best_fitness = -1;

  // allocate space for the population
  population = new unsigned int[psize];
  alive      = new bool[psize];
  fitnesses  = new int[psize];

  // randomly generate the initial population
  fprintf(stdout,"Initial population:\n");
  fprintf(stdout,"    Hex      Fitness\n");
  fprintf(stdout,"-----------  -------\n");
  for(i=0;i<psize;i++) {
    if(i<Np) {
      do {
        population[i] = rand32();
      } while(fitness(population[i])>max_init_fitness);

      fitnesses[i]  = fitness(population[i]);
      alive[i]      = true;

      reduction_or  = reduction_or  | population[i];
      reduction_and = reduction_and & population[i];

      // print initial population
      fprintf(stdout,"0x%08x      %i\n",population[i],fitnesses[i]);

    } else {
      alive[i]      = false;
      fitnesses[i]    = -1;
    }
  }

  fprintf(stdout,"\n");
  fprintf(stdout,"Reduction-And of initial population is: 0x%08x\n",reduction_and);
  fprintf(stdout,"Reduction-Or  of initial populaiton is: 0x%08x\n\n",reduction_or);

  bool mutant;
  // begin genetic loop
  for(int current_gen = 0; current_gen < Ng; current_gen++) {
    fprintf(stdout,"* Starting Generation %i *\n\n",current_gen+1);

    // BENE GESSERIT BREEDING PROGRAM
    fprintf(stdout,"Starting the Bene Gesserit Breeding Program:\n");
    fprintf(stdout,"Child  Parent A  Parent B  Value       Fitness\n");
    fprintf(stdout,"-----  --------  --------  ----------  -------\n");

    // figure out what the sum of the fitnesses of the living member is
    sum = 0;
    for(i=0;i<psize;i++) {
      if(alive[i]) sum+=fitnesses[i];
    }

    for(i=0;i<nkids;i++) {
      // select unique, living parents a and b
      do {
        kid = randOnRange(0,sum-1);
        running_sum = 0;
        for(a=0;a<psize;a++) {
          if(!alive[a]) continue;
          running_sum += fitnesses[a];
          if(running_sum>=kid) break;
        }

        if(a==psize) {
          fprintf(stderr,"DANGER WILL ROBINSON: A==PSIZE\n");
        }

        kid = randOnRange(0,sum-1);
        running_sum = 0;
        for(b=0;b<psize;b++) {
          if(!alive[b]) continue;
          running_sum += fitnesses[b];
          if(running_sum>=kid) break;
        }

        if(b==psize) {
          fprintf(stderr,"DANGER WILL ROBINSON: A==PSIZE\n");
        }

      } while(a==b);

      // find a spot for the kid -- open spots have -1 fitness
      for(kid=0;kid<psize;kid++) {
        if(fitnesses[kid]==-1) break;
      }

      if(kid==psize) {
        fprintf(stderr,"DANGER WILL ROBINSON!! No room for child in population\n");
      }

      // generate the kid
      population[kid] = crossover(population[a],population[b]);

      fprintf(stdout," %4i    %4i      %4i    ",
              i+1,a+1,b+1);

      // consider mutating the kid
      mutant = false;
      if( randNorm() < Rm ) {
        population[kid] = mutate(population[kid]);
        mutant = true;
      }

      fitnesses[kid]  = fitness(population[kid]);

      fprintf(stdout,"0x%08x   %i",population[kid],fitnesses[kid]);
      if(mutant) fprintf(stdout,"  \tMuTaNt");
      fprintf(stdout,"\n");

    }

    // now set all as alive
    for(i=0;i<psize;i++) alive[i] = true;

    // SURVIVAL OF THE FITTEST
    for(i=0;i<nkids;i++) {
      // kill the weakest living member
      b = 33; // b is "min_fitness"
      a = -1; // a is index of member with min fitness

      for(kid=0;kid<psize;kid++) {
        if(!alive[kid]) continue; // skip dead members
        if(fitnesses[kid]<b) {
          b = fitnesses[kid];
          a = kid;
        }
      }

      // now kill "a"
      alive[a] = false;
    }

    // print the population
    fprintf(stdout,"\n");
    fprintf(stdout,"Starting Survival of the Fittest:\n");
    fprintf(stdout,"Member   Value       Fitness\n");
      fprintf(stdout,"------   ----------  -------\n");
    for(i=0;i<psize;i++) {
      fprintf(stdout," %4i    0x%08x   %4i",i+1,population[i],fitnesses[i]);
      if(!alive[i]) fprintf(stdout,"     Killed");
      fprintf(stdout,"\n");
    
    }

    // set fitnesses of the dead to -1
    for(i=0;i<psize;i++) {
      if(alive[i]) continue;
      fitnesses[i] = -1;
    }

    fprintf(stdout,"\n");
  }

  fprintf(stdout,"** End of Genetic Algorithm **\n\n");

  best = 0;
  best_fitness = 0;
  for(i=0;i<psize;i++) {
    if(fitnesses[i]>best_fitness) {
      best_fitness = fitnesses[i];
      best = population[i];
    }
  }

  delete population;
  delete alive;
  delete fitnesses;

  return best;
}

int fitness(unsigned int x) {
  int count = 0;
  unsigned int mask = 0x00000001;
  for(int i=0;i<32;i++) {
    if((mask&x)!=0) count++;
    mask = mask << 1;
  }
  return count;
}

unsigned int mutate(unsigned int x) {
  int bit = randOnRange(0,31);
  unsigned int mask = 0x00000001;
  mask = mask << bit;

  return x ^ mask;
}

unsigned int crossover(unsigned int a, unsigned int b) {
  unsigned int mask = 0x00000001;
  unsigned int child = 0x00000000;

  for(int i=0;i<32;i++) {
    if(rand()%2==0) {
      child = child | ( mask & a );
    } else {
      child = child | ( mask & b );
    }
    mask = mask << 1;
  }

  return child;
}

